// add zero-padding methods

SoundFileLib {
	classvar <server, serverName, pathnames, <lib;
	classvar allPathnames, allLibs;
	*new { this.shouldNotImplement(thisMethod) }
	*initClass {
		Class.initClassTree(Server);
		allPathnames = IdentityDictionary[];
		allLibs = IdentityDictionary[];
		this.server_(Server.default);
	}
	*server_ { |argServer|
		server = argServer ? Server.default;
		serverName = server.name;
		// pathnames for all servers
		allPathnames.atFail(serverName, { allPathnames.add(serverName -> Set[]) });
		// properties lib for all servers
		allLibs.atFail(serverName, { allLibs.add(serverName -> IdentityDictionary[]) });
		pathnames = allPathnames[serverName]; // pathnames for current server
		lib = allLibs[serverName]; // properties lib for current server
	}
	load { |argPathnames, selector|
		var messages;
		
		messages = [];
		argPathnames.removeAllSuchThat({ |string| string.last == $/ }); // ignore directories
		
		if( server.serverRunning.not, {
			error("Can't load SoundFile if Server is not running.");
			this.halt;
		});

		argPathnames.do({ |path|
			var soundFile, bufnum, numChannels, properties, msg;
			
			soundFile = SoundFile.new;
			
			if( soundFile.openRead(path), {
				if( pathnames.includes(path), {
					error(path ++ " already loaded.")
				},{
					bufnum = server.bufferAllocator.alloc(1);
					numChannels = soundFile.numChannels;
					properties = [
						'bufnum' -> bufnum,
						'headerFormat' -> soundFile.headerFormat,
						'sampleFormat' -> soundFile.sampleFormat,
						'numFrames' -> soundFile.numFrames,
						'numChannels' -> numChannels,
						'sampleRate' -> soundFile.sampleRate,
						'duration' -> soundFile.duration
					];
					msg = this.perform(selector, bufnum, path, numChannels);
					messages = messages.addAll(msg);
					Routine.run {
						server.sync;
						server.listSendBundle(nil, messages);
					};
					lib.put(path.asSymbol, IdentityDictionary[].addAll(properties));
				});
			},{
				error(path ++ " could not be opened.");
				argPathnames.remove(path);
			});
			
			soundFile.close;
		});
		
		pathnames = pathnames.union(argPathnames);
		inform("loaded.");
	}
	allocReadMsg { |bufnum, path|
		^[["/b_allocRead", bufnum, path]]
	}
	allocMsg { |bufnum, path, chans|
		^[["/b_alloc", bufnum, 65536, chans], ["/b_read", bufnum, path, 0, 0, 0, 1]]
	}
	*load { |path|
		if( File.exists(path), { ^super.new.load([path], 'allocReadMsg');
		}, { error(path ++ " not found.") });
	}
	*loadAll { |pathnames|
		pathnames.do({ |path| this.load(path) });
	}
	*loadDir { |dir|
		if( File.exists(dir), { ^super.new.load((dir ++ "*").pathMatch, 'allocReadMsg');
		}, { error(dir ++ " not found.") });
	}
	*preLoad { |path|
		if( File.exists(path), { ^super.new.load([path], 'allocMsg');
		}, { error(path ++ " not found.") });
	}
	*preLoadAll { |pathnames|
		pathnames.do({ arg path; this.preLoad(path) });
	}
	*preLoadDir { |dir|
		if( File.exists(dir), { ^super.new.load((dir ++ "*").pathMatch, 'allocMsg');
		}, { error(dir ++ " not found.") });
	}
	*free { |path|
		var bufnum;
		if( pathnames.includes(path), { 
			bufnum = this.[path].['bufnum'];
			lib.removeAt(path.asSymbol);
			pathnames.remove(path);
			server.bufferAllocator.free(bufnum);
			server.sendMsg("/b_free", bufnum, ["/b_zero", bufnum]);
		},{ error(path ++ " is not in SoundFileLib.") });
	}
	*freeAll { |pathnames|
		pathnames.do({ |path| this.free(path) });
	}
	*freeDir { |dir|
		var bufnums, paths, messages;
		if( this.dirs.includes(dir), {
			bufnums = this.findBufnumsForDir(dir);
			paths = (dir ++ "*").pathMatch;
			paths.removeAllSuchThat({ |string| string.last == $/ });
			paths.do({ |path|
				server.bufferAllocator.free(this.[path].['bufnum']);
				lib.removeAt(path.asSymbol);
				pathnames.remove(path);
			});
			messages = bufnums.collect({ |num| ["/b_free", num, ["/b_zero", num]] });
			server.listSendBundle(nil, messages);
		},{ error(dir ++ " is not in SoundFileLib.") });
	}
	*clear { // free everything on current server
		var bufnums, messages;
		bufnums = this.bufnums;
		messages = bufnums.collect({ |num| ["/b_free", num, ["/b_zero", num]] });
		server.listSendBundle(nil, messages);
		bufnums.do({ |num| server.bufferAllocator.free(num) });
		allPathnames.add(serverName -> Set[]);
		allLibs.add(serverName -> IdentityDictionary[]);
		this.server_(server);
	}
	*at { |path|
		if( pathnames.includes(path), { 
			^lib[path.asSymbol];
		},{ error(path ++ " is not in SoundFileLib.") });
	}
	*pathnames {
		^pathnames.asArray.sort;
	}
	*filenames {
		^pathnames.collect({ |path| path.basename }).asArray.sort;
	}
	*dirs {
		^pathnames.collect({ |path| path.dirname ++ "/" }).asSet.asArray.sort;
	}
	*bufnums {
		^lib.values.collect({ arg properties; properties.['bufnum'] }).sort;
	}
	*findBufnumsForDir { |dir|
		var paths;
		paths = (dir ++ "*").pathMatch;
		paths.removeAllSuchThat({ arg string; string.last == $/ });
		^paths.collect({ arg path; this.[path].['bufnum'] });
	}
	*postContents {
		Post << "---------------------------------" << nl;
		Post << "SoundFileLib" << nl << nl;
		Post << "Server: " << serverName << nl << nl;
		if( lib.isEmpty, { inform("SoundFileLib is empty.");
		}, {
			this.pathnames.do({ |path| var properties;
				
				properties = lib.[path.asSymbol];
				
				Post << path << nl;
				Post.tab << "bufnum: " << properties['bufnum'] << nl;
				Post.tab << "headerFormat: " << properties['headerFormat'] << nl;
				Post.tab << "sampleFormat: " << properties['sampleFormat'] << nl;
				Post.tab << "numFrames: " << properties['numFrames'] << nl;
				Post.tab << "numChannels: " << properties['numChannels'] << nl;
				Post.tab << "sampleRate: " << properties['sampleRate'] << nl;
				Post.tab << "duration: " << properties['duration'] << nl << nl;
			});
		});
		Post << "---------------------------------" << nl;
	}
}

SFLib : SoundFileLib {}