/*
	Filename: VirtualRoom.sc 
	created: 21.4.2005 

	Copyright (C) IEM 2005, Christopher Frauenberger [frauenberger@iem.at] 

	This program is free software; you can redistribute it and/or 
	modify it under the terms of the GNU General Public License 
	as published by the Free Software Foundation; either version 2 
	of the License, or (at your option) any later version. 

	This program is distributed in the hope that it will be useful, 
	but WITHOUT ANY WARRANTY; without even the implied warranty of 
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
	GNU General Public License for more details. 

	You should have received a copy of the GNU General Public License 
	along with this program; if not, write to the Free Software 
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 

	IEM - Institute of Electronic Music and Acoustics, Graz 
	Inffeldgasse 10/3, 8010 Graz, Austria 
	http://iem.at
*/


/* 	NOTE: 
	the coordinate system is given according to the listener's head:
	x-axis (nose), y-axis (left-ear) and z-axis (vertex)
	
	* left-deep-corner			  *
		       x				  |
		       |	       		  |
		      / \		  		  depth
		y<---| z |			  |
		     ----				  |
							  |
	*---------- width ------------* (origin) 

	Rooms are defined by the origin and width/depth/height
*/


/* 	Class: VirtualEnvironment
	Provides a virtual room with sources, listener and binaural rendering
*/
VirtualRoom {
	
	// the path to Kemar
	classvar <>kemarPath = "KemarHRTF/";
	
	// the list of sources played
	classvar csources;

	// the node proxies
	classvar encoded, bin, out, revIn;

	// maximum of sources allowed (exclusive reflections) 
	classvar <>maxSources = 10;
	
	// roomProperties (reverberation, reflections)
	classvar roomProperties;
		
	// the room as [x, y, z, depth, width, height] where xyz is the origin
	var <>room;
	
	// the listener as NodePorxy.control with slots [ x, y, z, orientation]
	var <>listener;
	
	// the list of sources for each instance
	var <>isources;
	
	// play/pause flag
	var pl=false;
	
	/*	Class method: *new
		create an instance and initialise the instance's list of sources
	*/
	*new {
		^super.new.isources_(Dictionary.new);
	}

	/*	Class method: *init
		init the class and create the node proxies
	*/
	*init {

		// initialise the rendering
		BinAmbi3O.kemarPath = kemarPath;
		BinAmbi3O.init('1_4_7_4');
	 	
		// the list of sources played as Dictionary
		csources = Dictionary.new;
		
	}
	
	/* 	Class method: *initClass
		executed at compile time, create variable Dictionary plus defaults
	*/
	*initClass {
		roomProperties = Dictionary.new;
	}

	// access methods	
	*refGain_ { arg value;
		if(value.isNil) { roomProperties.removeAt(\refGain) } { roomProperties.put(\refGain, value) };
		csources.do({ |source| source.set(\refGain, value) });
	}
	*refGain { ^roomProperties.at(\refGain).value ?? 0.6 }
	*revGain_ { arg value;
		if(value.isNil) { roomProperties.removeAt(\revGain) } { roomProperties.put(\revGain, value) };
		out.set(\rG, value);
	}
	*revGain { ^roomProperties.at(\revGain).value ?? 0.1 }
	*revTime_ { arg value;
		if(value.isNil) { roomProperties.removeAt(\revTime) } { roomProperties.put(\revTime, value) };
		out.set(\rT, value);
	}
	*revTime { ^roomProperties.at(\revTime).value ?? 1 }
	*hfDamping_ { arg value;
		if(value.isNil) { roomProperties.removeAt(\hfDamping) } { roomProperties.put(\hfDamping, value) };
		out.set(\hD, value);
	}
	*hfDamping { ^roomProperties.at(\hfDamping).value ?? 0.6 }

	/* 	Class method: *gui
		provide a GUI for the room properties
	*/
	*gui {
		var w, f, s, v, t;
		var height = 15;
		s = Array.newClear(4);
		v = Array.newClear(4);
		roomProperties.put(\rTSpec, [0, 3].asSpec);			w = SCWindow("Virtual Room Properties", Rect(128, 64, 340, 150));
		w.view.decorator = f = FlowLayout(w.view.bounds,Point(4,4),Point(4,2));

		t = SCStaticText(w, Rect(0, 0, 75, height+2));
		t.string = "RefGain: ";
		v[0] = SCStaticText(w, Rect(0, 0, 30, height+2));
		s[0] = SCSlider(w, Rect(0, 0, 182, height));
		s[0].value = VirtualRoom.refGain;
		s[0].action = { 
			VirtualRoom.refGain = s[0].value;
			v[0].string = s[0].value.round(0.01).asString;
		};
		f.nextLine;
		
		t = SCStaticText(w, Rect(0, 0, 75, height+2));
		t.string = "RevGain: ";
		v[1] = SCStaticText(w, Rect(0, 0, 30, height+2));
		s[1] = SCSlider(w, Rect(0, 0, 182, height));
		s[1].value = VirtualRoom.revGain;
		s[1].action = { 
			VirtualRoom.revGain = s[1].value;
			v[1].string = s[1].value.round(0.01).asString;
		};
		f.nextLine;
		
		t = SCStaticText(w, Rect(0, 0, 75, height+2));
		t.string = "RevTime: ";
		v[2] = SCStaticText(w, Rect(0, 0, 30, height+2));
		s[2] = SCSlider(w, Rect(0, 0, 182, height));
		s[2].value = roomProperties.at(\rTSpec).unmap(VirtualRoom.revTime);
		s[2].action = { 
			var val = roomProperties.at(\rTSpec).map(s[2].value);
			VirtualRoom.revTime = val;
			v[2].string = val.round(0.1).asString;
		};
		f.nextLine;
		
		t = SCStaticText(w, Rect(0, 0, 75, height+2));
		t.string = "hfDamping: ";
		v[3] = SCStaticText(w, Rect(0, 0, 30, height+2));
		s[3] = SCSlider(w, Rect(0, 0, 182, height));
		s[3].value = VirtualRoom.hfDamping;
		s[3].action = { 
			VirtualRoom.hfDamping = s[3].value;
			v[3].string = s[3].value.round(0.01).asString;
		};
		f.nextLine;
		s.do({|x|x.action.value });
		
		w.front;
	}
	
	/*	Class method: *start
		start the rendering of the Virtual Room
	*/
	*start {

		// the whole rendering process encode - decode - convolution
		revIn = NodeProxy.audio(numChannels: 2);
		encoded = NodeProxy.audio(numChannels: 16);
		bin = NodeProxy.audio(numChannels: 2);
		bin.source = { BinAmbi3O.ar( encoded.ar ) };
		revIn.source = { bin.ar }; 	// not ideal...
		out = NodeProxy.audio(numChannels: 2);
		out.source = { arg rT = 1, rG = 0.1, hD = 0.6;
			bin.ar + (AdCVerb.ar( revIn.ar, revTime: rT, hfDamping: hD) * rG) 
		};
	
	}	
	/*	Class method: *start
		stop the rendering engine
	*/
	*stop {
		// free the node proxies...
		bin.free;
		revIn.free;
		encoded.free;
	
	}		
		
	/* 	Class method: update
		rebuilds the encoded sources from the csources list
	*/
	*update {	
		csources.postln;	
		if (csources.size != 0, 
			// insert all sources and sum them
			{ encoded.source = csources.sum; }, 
			{ encoded.source = nil });
		// make sure refGain is set properly - not necessary??
		// csources.do({ |source| source.set(\refGain, VirtualRoom.refGain) });
	}

	/* 	Class method: ar
	   	Audio rate method
		Return
			binaural output
	*/
	*ar {  
		^out.ar;
	}
	
	/* 	method: play
		start playing the virtual room
		add the sources from the instance list to the class's list and make them play
	*/
	play { 
		pl = true;
		isources.keys.do({ |key|
			csources.put(this.identityHash.asString ++ key, isources[key]);
		});
		// update the encoding node proxy
		VirtualRoom.update;
	}
	
	/* 	method: stop
		stop playing the virtual room
		remove the sources from the class's list and make them play
	*/
	stop { 
		pl = false;
		isources.keys.do({ |key|
			("removing key " ++ key).postln;
			csources.removeAt(this.identityHash.asString ++ key);
		});
		// update the encoding node proxy
		VirtualRoom.update;
	}

	/* 	method: addSourceLight
		add a source to the virtual room, but just use 4 reflections. This is though for using many sources 
		Parameter:
			key A string to identify the source
			source A mono sound source as NodeProxy.audio
			pos A NodeProxy.control with [xpos, ypos, zpos]
	*/
	addSourceLight { arg source, key, pos;
	
		var  synthFunc; 
		
		("VirtualRoom: adding source " ++ key ++ " - the light way...").postln;
		
		synthFunc = { arg refGain = 0;
			var sourcePositions;
			var distances, gains, phis, thetas, delTimes, gainSource;
			var xpos, ypos, zpos, lx, ly, lz, lo;
			
			xpos = pos.kr(1,0); ypos = pos.kr(1,1); zpos = pos.kr(1,2);
			lx = listener.kr(1,0); ly = listener.kr(1,1); lz = listener.kr(1,2); lo = listener.kr(1,3);
			
			// direct source + 4 reflections		
			sourcePositions = [
				xpos, ypos, zpos, 
				this.room[0] + (2 * this.room[3]) - xpos, ypos, zpos, 
				this.room[0] - xpos, ypos, zpos, 
				xpos, this.room[1] + (2 * this.room[4]) - ypos, zpos, 
				xpos, this.room[1] -ypos, zpos
			];
			#phis, thetas, distances = sourcePositions.clump(3).collect({ |pos|
				var planeDist;
				planeDist = hypot(pos[1]-ly, pos[0]-lx);
				[atan2(ly-pos[1], pos[0]-lx) + lo, atan2(pos[2]-lz, planeDist), hypot(planeDist, lz - pos[2])];
			}).flop;		

			delTimes = ( distances / 340 );
			gains = (distances + 1).reciprocal; 
			(1..4).do({ | i | gains[i] = gains[i] * refGain });
		
			// sum up the encoded channels of all sources (original + reflections) 
			// DelayL replacement with BufRead....
			DelayL.ar( source.ar, 2, delTimes, gains.squared).collect( { |ch, i| 
				PanAmbi3O.ar(ch, phis[i], thetas[i]); }).sum;
		};

		// produce the NodeProxy.audio
		isources.put(key, NodeProxy.audio(numChannels: 16));
		isources[key].set(\refGain, VirtualRoom.refGain);
		isources[key].source = synthFunc; 
		
		// update the encoding node proxy
		VirtualRoom.update;
		if (pl, { this.play; } );
	}


	/* 	method: addSource
		add a source to the virtual room 
		Parameter:
			key A string to identify the source
			source A mono sound source as NodeProxy.audio
			pos A NodeProxy.control with [xpos, ypos, zpos]
	*/
	addSource { arg source, key, pos;
		
		var  synthFunc; 
		
		("VirtualRoom: adding source " ++ key).postln;
		
		synthFunc = { arg refGain;
			var phi, theta, planeDist, roomDist, refGain2; 
			var roomModel, sourceAndRefs; 
			var distances, gains, phis, thetas, delTimes, gainSource;
			var xpos, ypos, zpos, lx, ly, lz, lo;
			refGain2 =  refGain.squared;
			
			xpos = pos.kr(1,0); ypos = pos.kr(1,1); zpos = pos.kr(1,2);
			lx = listener.kr(1,0); ly = listener.kr(1,1); lz = listener.kr(1,2); lo = listener.kr(1,3);
							
			// calculate angles for source
			planeDist = hypot(ypos-ly, xpos-lx);
			phi = atan2(ly-ypos, xpos-lx) + lo;
			theta = atan2(zpos-lz, planeDist);
			roomDist = hypot(planeDist, lz - zpos);
			// calculate the room model
			roomModel = Room3D.new;
			roomModel.room = this.room;
			sourceAndRefs = [phi, theta, roomDist] ++ 
				roomModel.refs10polar(xpos, ypos, zpos, lx, ly, lz);
		
			#phis, thetas, distances = sourceAndRefs.clump(3).flop;
			delTimes = ( distances / 340 );
			gains = (distances + 1).reciprocal; 
			(1..10).do({ | i | gains[i] = gains[i] * if (i.inclusivelyBetween(5,8), refGain, refGain2) });
		
			// sum up the encoded channels of all sources (original + reflections) 
			// DelayL replacement with BufRead....
			DelayL.ar( source.ar, 2, delTimes, gains.squared).collect( { |ch, i| 
				PanAmbi3O.ar(ch, phis[i], thetas[i]); }).sum;
		};

		// produce the NodeProxy.audio
		isources.put(key, NodeProxy.audio(numChannels: 16));
		isources[key].set(\refGain, VirtualRoom.refGain);
		isources[key].source = synthFunc; 
		
		// update the encoding node proxy
		VirtualRoom.update;
		if (pl, { this.play; } );
	}
	
	
	/* 	method: removeSource
		remove a source to the virtual room 
		Parameter:
			key The key to identify the source
	*/
	removeSource { arg key;
	
		// free the NodeProxy and remove it from the instance dict
		isources[key].free; isources.removeAt(key);
		
		// remove it also from csources when playing the room (wouldnt do anything if key doesnt exist)
		csources.removeAt(this.identityHash.asString ++ key);

		// update the encoding node proxy
		VirtualRoom.update;
		if (pl, { this.play; } );
	}
	
}	

