/*

CV's represent a single float or an array of floats that all have the same range of values (determined by a ControlSpec).
CV's provide the "glue" that allows a value to be easily connected to MIDI, GUIs, Nodes, Buses, etc.
CV responds to value and value_(v) by getting or setting its value (constrained to the range defined 
by its ControlSpec). They respond to input and input_ by getting or setting a value in the range [0,1] 
that maps (via its ControlSpec) to the CV's current value.

CVEvent groups a collection of CVs together, they are stored as key/value pairs within the event 
and are identified in an array of keys stored in the instance variable valueItems.

CVevent responds to value and input by collecting the value or input of the identified CVs
and responds to value_(arrayOfValues) and input_(arrayOfInputs) by assigning elements of the argument
array to the corresponding CVs:

CVEvent-value: valueItems.collect { | key | this[key].value } 

CVEvent-value_ : { | vs|  vs.do { | v, i| this[valueItems[i] ].value_(v) } }
 		
A CVEvent can consist of CVs, CVEvents, or any other object that responds to value, input, value_, input_.
*/
CVEvent : Event {
	var <>name;
	var <path;
	var <>valueItems;
	
	*new { | ...args |
		^super.new.init(args);
	}
	
	init { | args |
		valueItems = args.collect({ | c | this.put(c.name, c); c.name });
	}

	*make { arg func; 
		var obj, args, names;
		obj = this.new;
		^obj.make(func)
	}
	
	make { arg func; 
		var obj, args, names;
		#args, names = this.makeArgs(func);
		valueItems = valueItems ++ names;
		super.make({func.valueArray(this, args)});
	}

	makeArgs { arg func;
		var argList, size, names, argNames;
		var theClassName, name, obj;
		
		size = func.def.argNames.size;
		argList = Array(size);
		argNames = Array(size);
		names = func.def.argNames;
		// first arg is Event under constructions, subsequent are CVs or instances of other classes
		if (size > 1, {
			1.forBy(size - 1, 1, { arg i;
				name = names.at(i).asSymbol;
				argNames = argNames.add(name);
				theClassName = func.def.prototypeFrame.at(i) ? \CV;
				obj = theClassName.asClass.new;
				this.put(name,obj);
				argList = argList.add(obj);
			});
		});
		^[argList, argNames];
		
	}
		
	value  { ^valueItems.collect { | n| this[n].value } }
	value_ { | vs|  vs.do { | v, i| this[valueItems[i] ].value_(v) } } 
	
	input  { ^valueItems.collect { | n| this[n].input } }
	input_ {| vs|  vs.do { | v, i| this[valueItems[i]].input_(v) } }
	
	path_ { | p |
		path = p;
		this.load
	}

	load { | argPath |
		var obj;
		path = argPath ? path;
		obj = this.getFile(path);
		this.setKeysValues(obj.interpret); 	
	}
	
	save { | path |
		var vals;
		vals = valueItems.collect { | name | [name, this[name].value] };
		this.putFile(this.keysValues.asCompileString, path)
	}
	
	keysValues { ^valueItems.collect { | name | [name, this[name].value] }; }
	
	setKeysValues { | kvArray | 
		var cv;
		kvArray.do { | v | 
			cv = this[v[0]];
			if (cv.notNil) 
				{ cv.value_(v[1]) } 
				{( "warning: " + v[0].asString + "is not defined in Conductor" + name).postln; ^nil; }
		}
	}

	
	getFile { arg argPath; var file, contents;
		if (File.exists(argPath)) {
			path = argPath;
			file = File(path,"r"); 
			contents = file.readAllStringRTF;
			file.close;
			^contents
		} {
			(argPath + "not found").postln;
			^nil
		}
	}
	
	putFile { | vals, argPath | 
		path = argPath ? path;
		File(path,"w").putAll(vals).close;
	}

}

/*
ConductorEvent is a CVEvent that can play/stop/pause/resume any collection of 
tasks. patterns. groups, synths, bus allocations, and buffers. 

*/


ConductorEvent : CVEvent {
	var <>server;
	var <>players;
	
	init { | args |
		server = Server.default;
		players = [];
		name = ">";
		super.init(args);
	}
		
	changed { | what, val |
		defer { super.changed(what, val) }
	}
	
	stop {
		players.do(_.stop);
		CmdPeriod.remove(this); 
			this.changed(\synch, 0);
	}
	
	play { 
		server.makeBundle(server.latency, {
			this.use{ players.do(_.play) };
		});
		CmdPeriod.add(this);  
		this.changed(\synch, 1);
	}

	pause { 
		players.do(_.pause);
		this.changed(\synch, 2);
	}
	resume { 
		players.do(_.resume);
		this.changed(\synch, 1);
	}

	cmdPeriod { 		
		 CmdPeriod.remove(this); 
		 this.changed(\synch, 0);  
	}
	
	add { | obj |
		players = players.add(obj);
		
	}
	
	remove { |obj|
		players.remove(obj)
	}
	
	action_ { | playFunc, stopFunc, pauseFunc, resumeFunc |
		this.add ( ActionPlayer(playFunc, stopFunc, pauseFunc, resumeFunc ) )
	}

	buffer_ { | ev| 
		ev.parent = CV.bufferEvent;
		this.add(ev);
	}
	
	controlBus_ { | ev, cvs|
		ev[\cvs] = cvs;
		ev.parent = CV.controlBusEvent;
		this.add(ev)
	}

	synth_ { | ev, cvs|
		ev[\cvs] = cvs;
		ev.parent = CV.synthEvent;
		this.add(ev)
	}

	synthDef_ { | function, cvs, ev|
		var name;
		name = function.hash.asString;
		SynthDef(name, function).store;
		ev = ev ? ();
		ev	.put(\instrument, name)
			.put(\cvs, cvs);
		ev.parent_(CV.synthEvent);
		this.add(ev);
		^ev
	}

	group_ { | ev, cvs|
		ev[\cvs] = cvs;
		ev.parent = CV.groupEvent;
		this.add(ev)
	}
	
	// TaskWithStreamKill handles sleeping on a signal correctly and provides clock, quant option
	task_ { |func, clock, quant|
		this.add(TaskWithStreamKill(func,clock, quant));
	}
	
	pattern_ { | pat, clock, event, quant |
		this.add(PatternPlayer(pat, clock, event, quant) )
	}


}

Conductor : ConductorEvent {
	classvar <defaultGUIFunction;
	var <>guiFunction;
	var <>guiItems;
	var <>guiFunctions;
	var <>stopOnClose;
	
	*initClass {
		defaultGUIFunction = #{ |win, name, conductor|
			conductor.drawPlayPause(win, name);
			conductor.drawItems(win, conductor.guiItems)
		}
	}
	init { arg args;
		super.init(args);
		guiItems = valueItems;
		guiFunctions = ();
		this.parent = GUIEvent.osx;
		stopOnClose = true;
		name = ">";
		guiFunction = defaultGUIFunction;
	}

	make { arg func; 
		var obj, args, names;
		#args, names = this.makeArgs(func);
		guiItems = guiItems ++ names;
		valueItems = guiItems;
		super.make({func.valueArray(this, args)});
	}
	
		
	show { arg argName, x = 128, y = 64, w = 900, h = 160;
		var win, bounds;
		
		win = SCWindow(argName ? "", Rect(x,y,w,h));
		win.view.decorator = FlowLayout(win.view.bounds);
		bounds = win.bounds;

		SimpleController(win).put(\close, { this.remove });

		this.draw(win);
		bounds.height = win.view.decorator.top + 16;
		win.bounds_(bounds);
		win.front;
		^win;
	}
	
	remove {			
		if (stopOnClose == true, { 		// so deferred synch messages
			this.stop; 				// don't update deleted
		});							// dependants
	}
	
	drawPlayPause { arg win, name; 
		if (players.size > 0) {
			PlayPause(win, Rect(0,0,50,20)).string_(name ? ">").model_( this); 
		};
		if (this[\settings].notNil) {
			this[\settings].draw(win, name, this[\settings])
		};			
		win.view.decorator.nextLine;
	}

	drawItems { | win, itemKeys |
		var guiFunction, cv;
		itemKeys.do{arg v;
			v.asArray.do{ arg v; 
				if ( (guiFunction = guiFunctions[v]).isNil) {
					cv = currentEnvironment[v];
					cv.draw(win, v, cv) 
				} {
					if (guiFunction.isKindOf(Array)) {
						#guiFunction...cv = guiFunction;
					} {
						cv = currentEnvironment[v]
					};
					guiFunction.value(win, v, cv)
				};
				v.update(this,\synch); 
			};						
			win.view.decorator.nextLine;
		};
	}
	
	draw { arg win, argName;
		this.use { 
			guiFunction.value(win, argName ? name, this)
		}
	}
	
	asArray {^[this]}	

	useSettings {
		this[\settings] = Settings(this);
		valueItems = valueItems.add(\settings);
	}

	usePresets {
		this[\settings] = Presets(valueItems, this);
		valueItems = valueItems.add(\settings);
	}
	
	useInterpolator {
		this[\settings] = Interpolator(valueItems, this);
		valueItems = valueItems.add(\settings);
	}

}


/*

(a = Conductor.make{|d, a, b, c, e = \Conductor|
	d.name_(\ff);
	a	.sp(200,100,5000);
	d.action_({ currentEnvironment.postln });
	e.make { |d, a, b, c| };

	d.synth_( (),[freq: a, amp: b] );
	e.pattern_(Pbind(\dur, Pwhite(0,0.001) + c, \freq, a, \amp, b));
	d.guiItems = #[a,b,c, e];
};

a.name_(\test);
a.show;
)
b = Conductor(a);
//b.show;

b.show
)


//
//	{ |win, key |
//		EZSlider(win, ~sliderdim ? 600@20, key, ~controlspecs[key]).connect(~key, win);
//	}
//	
//	{ |win, key |
//		SCSlider(win, ~sliderdim ? 600@20, key, ~controlspecs[key]).connect(~key, win);
//	}

(
a = Conductor.make{ |d, a, b, c|
	var buf;
	a	.sp((0..511)/512, -1,1);
	d.action_(
		{  ~buf = buf = Buffer.alloc(Server.default, a.value.size) },
		{  ~buf.free }
	); 

	~guiFunctions[\a] = { |win, name| 
		var labelView, slider, w;
		w = SCVLayoutView(win, Rect(0,0,80, 200));
		labelView = SCStaticText(w, Rect(0,0,80, 20));
		labelView.string = name;
		labelView.align = \right;
		labelView = SCButton(w, Rect(0,0,80, 85) )
			.states_([ ["send buffer\n", Color.black,Color.gray(1,0)] ])
			.action_({ a.input_(slider.value); buf.setn(0, a.value) });
		labelView = SCButton(w, Rect(0,0,80, 85) )
			.states_([ ["get buffer\n", Color.black,Color.gray(1,0)] ])
			.action_({ buf.getn(0, a.value.size, { |vals| a.value = vals}) });
		
		slider = SCMultiSliderView(win, Rect(0,0,520,200)).connect(a);
		slider.thumbSize = 520/a.value.size;
		slider.valueThumbSize_(2.0);
		d[\slider] = slider;
	};
	~guiFunctions[\b] = { |win, name| EZSlider(win, 600@20, name, b.spec).connect(b) };
	
};
a.show;

)
CV
*/