/*
	Utilities to connect ControlValues to Node controls, Control Buses, and Node Proxys.
interface:
	Array-connectToBus	(server, index)	
		the array consists of SimpleNumbers and ControlValues
		
	Array-connectToNode (server, nodeID)
	Array-connectToNodeProxy (nodeProxy)
		the array consists of labels alternating with a SimpleNumber or ControlValues
		

	Bus-setControls( arrayOfValuesAndControlValues)
	Node-setControls( [name, (control)Value, name, (control)Value...])
	NodeProxy-setControls( [name, (control)Value, name, (control)Value...])
		'setControls' acts like 'set', but replaces ControlValues with their
		values in the array and creates the needed synchronization logic (see below).
	
implementation:
A SimpleController dependant on the ControlValue relays changes to the NC, CB, or NP.
OSCresponders are used to remove the SimpleController when the NC, CB or NP is freed.
This is implemented with a dummy OSC message when freeing a Control Bus.

*/

+Object {
	isSymbol { ^false }
}

+Symbol {
	isSymbol { ^true }
}

+Server {
	removeBusLinks { arg busIndex;
			OSCresponder.respond(0, this.addr, ["/c_end", busIndex]);
	}

	removeControlLinks { arg nodeID;
			OSCresponder.respond(0, this.addr, ["/n_end", nodeID]);
	}

}

+Bus {
	free {
		if(index.isNil,{ (this.asString + " has already been freed").warn; ^this });
		if(rate == \audio,{
			server.audioBusAllocator.free(index);
		},{
			server.removeBusLinks(index);
			server.controlBusAllocator.free(index);
		});
		index = nil;
		numChannels = nil;
	}
	makeControls { arg arrayOfControlValues, size;
			^(["/c_setn", index] ++ size ++ arrayOfControlValues.connectToBus(server,index, size)); 
	}
	setControls { arg arrayOfControlValues, size;
		server.sendBundle(server.latency, ["/c_setn", index] ++ size ++ 			arrayOfControlValues.connectToBus(server,index, size).flatten(1))
	}
	
	*controls { arg args, server, size = 1;
		var bus;
		bus = Bus.control(server, size);
		bus.setControls(args, size);
		^bus;
	}
}

+Array {
	
	connectToNode { arg server, nodeID;
		var msg, parameters, nodeLinks, name, cv, modifier, size;
		size = this.size;
		parameters = this.copy;
		nodeLinks = Array(size);

		forBy (1, size - 1, 2, { arg i;
			cv  = parameters.at(i);
			if (cv.isKindOf(Function)) { cv = cv.value; parameters.put(i,cv)};
			if( (cv.isNumber or: cv.isSymbol).not, { 
				msg = ['/n_set', nodeID, parameters.at(i - 1)];
				if (cv.isSequenceableCollection) {	
					#cv, modifier = cv;
					cv.asArray.do({ arg cv; nodeLinks.add( cv.connectToServer(server, msg, modifier) ) });
					parameters.put(i, modifier.value(cv));
				} {			
					nodeLinks.add( cv.connectToServer(server, msg, cv) );
					parameters.put(i, cv.value);
				};
			});
		});	
			
		if (nodeLinks.size > 0, { 
			OSCpathResponder(server.addr, ["/n_end", nodeID], 
				{ arg time, resp, msg; nodeLinks.do({ arg n; n.remove}); resp.remove;} 
			).add;
		});
		
		^parameters;
	}

	connectToBus { arg server, busIndex, size;
		var busLinks, parameters, modifier, msg;
		parameters = this.copy;
		busLinks = Array(this.size);
		parameters.do({ arg cv, i;  

			if( (cv.isNumber).not, { 
//				msg = ['/c_setn', busIndex + i, size];
				msg = ['/c_set', busIndex + i];
				if (cv.isSequenceableCollection) {	
					#cv, modifier = cv;
					cv.asArray.do({ arg cv; busLinks.add(cv.connectToServer(server,msg, modifier)) }); 
					parameters.put(i, modifier.value(cv));
				} {			
					busLinks.add(cv.connectToServer(server, msg, cv)); 
					parameters.put(i, cv.value);
				};
			});
		});

		if (busLinks.size > 0, { 
			// this is a dummy OSC command, used by Server-removeBusLinks
			// Bus objects use this when freed
			// If using raw OSC commands, user must call server.removeBusLinks(busIndex)
			// to remove
			OSCpathResponder(server.addr, ['/c_end', busIndex], 
				{ arg time, resp, msg;
					busLinks.do({ arg n; n.remove});
					resp.remove;
				 } ).add;
		});
		^parameters;
	}
	connectToNodeProxy { arg np;
		var parameters, nodeLinks, name, cv, modifier,size;
		size = this.size;
		parameters = this.copy;
		nodeLinks = Array(size);

		forBy (1, size - 1, 2, { arg i;
			cv  = parameters.at(i);
			if( (cv.isNumber).not, { 
				name = parameters.at(i - 1);
				if (cv.isSequenceableCollection) {	
					#cv, modifier = cv;
					cv.asArray.do({ arg cv; nodeLinks.add( cv.connectToNodeProxy(np, name, modifier) ) });
					parameters.put(i, modifier.value(cv));
				} {			
					nodeLinks.add( cv.connectToNodeProxy(np, name, cv) );
					parameters.put(i, cv.value);
				};
			});
		});
		if (nodeLinks.size > 0, { 
			OSCpathResponder(np.group.server.addr, ["/n_end", np.group.nodeID], 
				{ arg time, resp, msg; nodeLinks.do({ arg n; n.remove}); resp.remove;} 
			).add;
		});
		^parameters;
	}

}

