/* 
 ClientNode and ControlValue are "models".
 Control sources use value_ and input_ to alter that state of a ClientNode or ControlValue
 Control sources that can be synchronized with the state of the ClientNode or ControlValue are
 also dependants that respond to 'synch' by synchronizing their state to that of their model.
 (The reasoning here is that the model may not go to the precise value 'suggested' by the control source. For example, a gui slider's resolution can be finer than the step size allowed by the ControlValue)

Base class for things that are controlled by client side groups
	start stop pause resume		standard control interface
	/synch					message that can update dependants such as GUI displays
	value_ 					provides a single method for start/stop/pause/resume
	input_					used for connecting to external control sources such as MIDI
*/
ClientNode {
	classvar <>addActions;
		
	var	<value;			// state of node: 0 - stopped, 1 - playing, 2 - paused
	var <>group;			// this is a ClientGroup or ServerGroup
	var <>target;			// Not everything sends to the server
	var <>addAction;		// but everything might send to the server...
	var <>name;

	*initClass { 
		addActions = IdentityDictionary.new;
		addActions.addAll([
			\addToHead -> 0,
			\addToTail -> 1,
			\addBefore -> 2,
			\addAfter -> 3
		]);
	}
	
	draw { arg layout;
		^this.defaultGUI.draw(layout)
	}
		
	defaultGUI { 	^PlayPauseStopButton(this).name_(name) }
	play {
		if (group.isNil, { 
			CmdPeriod.add(this) 
		}, {
			group.children = group.children.add(this)
		});
		value = 1;
		this.changed(\synch);
	}
	
	stop { 
		if (group.isNil, { CmdPeriod.remove(this) });	
		value = 0;
		this.changed(\synch);
	 }
	
	pause { 
		value = 2;
		this.changed(\synch);
	 }
	
	resume {  
		value = 1;
		this.changed(\synch);
	}	
			
	value_ { arg newStatus;
		newStatus = (newStatus max: 0 min: 3).asInteger;
		this.changed(\action);
		if (newStatus != value, {
			[  	{ this.stop(true); }, 
				{ if (value == 0, {this.play(true)}, {this.resume(true) }) }, 
				{ this.pause(true) },
				{ this.resume(true) }
			].at(newStatus.asInteger).value;
		}, {
	//		"redundant synch message".warn;
		}); 
	}
	
	input_ { arg v; this.value_(v) }
	input { ^value }

	cmdPeriod {
		this.stop(false); 
	}
	
	server { ^target.server }
	
	asGroup { ^target.asTarget.asGroup }
	
	asNodeID { ^target.asTarget.nodeID }
	
	isPlaying { ^value != 0 }
	isRunning { ^value != 0 }
	
}

ClientGroup : ClientNode {
	var <>childDefinitions, <>children;

	*new { arg target, addAction;
		^super.newCopyArgs(0, nil,target, addAction ? \addToTail);
	}
	
	play { arg flag = true; 
	 	super.play;
	 	childDefinitions.do({ arg c; c.play(flag) });
	 }
	 
	stop { arg flag = true;
	 	super.stop;
	 	children.do({ arg c; c.stop(flag) });
	 	children = nil;
	}
		 	
	pause { arg flag = true;
	 	super.pause;
	 	children.do({ arg c; c.pause(flag) });
	}	 	
	resume { arg flag = true;
	 	super.resume;
	 	children.do({ arg c; c.resume(flag) });
	}
	
	add { arg obj;  
		obj.asArray.do({ arg g; 
			g.group_(this); 
			if (value == 1, { g.play });
		});		
	}	 	

	attach { arg obj;  
		obj.asArray.do({ arg g; g.group_(this); 
			childDefinitions = childDefinitions.add(g);
			if (value == 1, { g.play });
		});		
	} 
	
	pattern_ 	{ arg pattern, event, clock, quant, target, addAction;
		var o; 
		this.attach(o = PatPlayer(pattern, event, clock, quant, target, addAction)); 
		^o
	}
	task_ 	{ arg func, clock, quant, target, addAction; 
		var o; 
		this.attach(o = TaskPlayer(func, clock, quant, target, addAction)); 
		^o
	}
	nodeProxy_ { arg func, args, clock, quant, target, addAction; 
		var o; 
		this.attach(o = NodeProxyPlayer(func, args, clock, quant, target, addAction)); 
		^o
	}
	synth_ 	{ arg name, args, target, addAction; 
		var o; 
		this.attach(o = ServerSynth(name, args, target, addAction)); 
		^o
	}
	synthDef_{ arg obj, args, target, addAction; 
		var o; 
		this.attach(o = ServerSynthDef(obj, args, target, addAction)); 
		^o
	}
	
}



/*
t.task.play(AppClock, true, 0.01)
t.task.stop
z = ClientGroup.new;

z.add(t = TaskPlayer({
	a = NodeProxy.audio(Server.local, 2).group_(z.asGroup);
	a.source = { arg f = 400, a = 0.1; SinOsc.ar(f,0,a)};
	a.play(group: z.asGroup);
	1.wait }, AppClock) );
z.play
z.stop
z.value_(0)
Tnodz.stop
a = ClientGroup.new;
b = ServerSynth("bubbles");
z.add(b);
a.play
a.pause
a.resume
a.stop
b.play
b.stop

Display.make({ arg d, gp = \ClientGroup;
	gp.add(ServerSynth("bubbles") );
	gp.add(NodeProxyPlayer({ arg n; n.source = {SinOsc.ar(400,0,0.1)}; }) );
	d.gui = [ControlButton(gp).name_("theGroup")];
	g = gp;
}).show;
Display.make({ arg d, gp = \ServerGroup, f, a;
	f.spec_(\freq);
//	gp.add(ServerSynth("bubbles") );
//	gp.add(NodeProxyPlayer({ arg n; n.source = {SinOsc.ar(400,0,0.1)}; }) );
	gp.add(PatPlayer(Pbind(\freq,Pseq(Array.series(10,100,100),inf) ,\dur,0.1), quant: 0.001) );
//	gp.add(ServerSynthDef({arg a = 0.1; SinOsc.ar(1000,0,a)}, [\a,a]));
	d.gui = [ [
				ControlButton(PlayPause(gp)).name_(" > ").states_([ [" >", Color.green,Color(0.2,0.2,0.2,1)],[" ||", Color.yellow,Color(0.2,0.2,0.2,1)] ]),
				ControlButton(Stop(gp)).name_(" = ").states_([ [" []", Color.red,Color(0.2,0.2,0.2,1)],[" []", Color.red,Color(0.2,0.2,0.2,1)] ])
			],
			 NSlider(f), NSlider(a)];
	g = gp;
}).show;

Display.make({ arg d, gp = \ClientGroup, f, a;
	f.spec_(\freq);
	gp.add(ServerSynth("bubbles") );
	gp.add(NodeProxyPlayer({ arg n; n.source = {SinOsc.ar(400,0,0.1)}; }) );
	d.add(PatPlayer(Pbind(\freq,Pseq(Array.series(10,100,100),inf) ,\dur,0.1), clock: SystemClock, quant: 1) );
//	gp.add(ServerSynthDef({arg a = 0.1; SinOsc.ar(1000,0,a)}, [\a,a]));
	g = gp;
}).show;
*/

