/*
 ControlEnvironment combines a Clientgroup capable of stopping and starting
 arbitrary collections of Nodes and PauseStreams with an environment that can be used
 to identify ControlValues, Clientgroups, Servergroups and other objects useful 
 for creating music.
 
*/

ControlEnvironment : Event {
	var <>clientgroup;
	var <>valueItems;
	var <>presets;
	var <filePath;
	var <>interpolationBase;
	var <>interpolationTarget;
	
	*new { | ...args |
		^super.new.init(args);
	}
	
	init { | args |
		clientgroup = ClientGroup.new;
		args.do({ | c | this.put(c.name, c) });
		valueItems = args;
	}
		
	*make { arg func; 
		^this.new.make(func);
	}
	
	make { arg func;
		var args;
		args = this.makeArgs(func);
		this.installMakeFuncArgs(args);
		super.make({ func.valueArray(this, args) });
	}
	
	installMakeFuncArgs{ arg func;
		var args;
		valueItems = valueItems ++ args;
	}

	makeArgs { arg func;
		var argList, size, names;
		var theClassName, name, obj;
		
		size = func.def.argNames.size;
		argList = Array(size);
		
		// other args are ControlValues or instances of other classes
		if (size > 1, {
			1.forBy(size - 1, 1, { arg i;
				name = func.def.argNames.at(i).asSymbol;
				theClassName = func.def.prototypeFrame.at(i) ? \ControlValue;
				obj = theClassName.asClass.new.name_(name);

				this.put(name, obj);
				argList = argList.add(obj);
			});
		});
		^argList;
		
	}
	
	name { ^clientgroup.name }	
	name_ { arg n; clientgroup.name = n }
	
// ControlValue style interface

	input { ^valueItems.collect({ arg cv; cv.input}) }
	input_{ arg input;  input.do({ arg v, i; valueItems.at(i).input_(v)}) }
	value { ^valueItems.collect({ arg cv; cv.value}) }
	value_{ arg input;  input.do({ arg v, i; valueItems.at(i).value_(v)}) }
	next { ^valueItems.collect({ arg cv; cv.value}) }
	
	
// Clientgroup style interface
	play { arg flag = true; clientgroup.play(flag) }
	stop { arg flag = true; clientgroup.stop(flag) }
	resume { arg flag = true; clientgroup.resume(flag) }
	pause { arg flag = true; clientgroup.pause(flag) }
	add { arg obj; clientgroup.add(obj) }
	attach { arg obj; clientgroup.attach(obj) }
   	pattern_ { arg pattern, event, clock, quant, target, addAction;
		clientgroup.pattern_(pattern, event, clock, quant, target, addAction)
	}
	task_ { arg func, clock, quant, target, addAction; 
		clientgroup.task_(func, clock, quant, target, addAction)
	}
	nodeProxy_ { arg func, args, clock, quant, target, addAction; 
		clientgroup.nodeProxy_( func, args, clock, quant, target, addAction );
	}
	synth_ { arg name, args, target, addAction; 
		clientgroup.synth_(name, args, target, addAction)
	}
	synthDef_ { arg obj, args, target, addAction; 
		clientgroup.synthDef_( obj, args, target, addAction)
	}
	group { arg newGp; ^clientgroup.group }
	group_ { arg newGp; clientgroup.group_(newGp) }
	
	asArray { 
		^[this] 			
	}

	addPreset { 	
		presets = presets.add(this.input); 
	}
	
	removePreset { arg index;
		if (presets.notNil, {
			presets.removeAt(index);
		});
	}

	setPreset { arg index;
		this.input_(presets.at(index));
	}
	
	setInterpolationTarget { arg index;
		var currentGoal, currentPreset;
		currentGoal = presets.at(index);
		currentPreset = this.input;
		interpolationBase = 2 * currentPreset - currentGoal;
		interpolationTarget = currentGoal - currentPreset 
		
	}

	interpolate { arg index;
		if (interpolationTarget.notNil, {
			this.input_( (index + 1) * interpolationTarget + interpolationBase);
		})
	}	

	load { arg path; var file;
		`path = path ? filePath;
		if (File.exists(path), {
			filePath = path;
			file = File(path,"r"); 
			this.presets_(file.readAllString.interpret);
			file.close;
		})
	}
	
	save { arg path; 
		filePath = path ? filePath;
		File(path,"w").putAll(presets.asCompileString).close;
	}
	

}

