
// AdhocClass: a flexible environment that can be evaluated like a stream to return SequenceNotes
// all object properties and functions (evaluation, new object generation, self-modification)
// must be stored in environment variables

// certain variables must have specific names to be found easily from outside:
// ~next = func to generate next value

AdhocClass {
	var	<>env;	// I'm providing a setter for env, but it's really for internal use only
				// be careful if you muck around with it!
	
	*new { arg initFunc, env, parentKeys;
			// the init func is evaluated in Environment.make
			// and should include setting of all env variables
			// env allows passing of a default environment
			// parentKeys says which data keys should be moved into the parent for inheritance
		^super.new.init(initFunc, env, parentKeys);
	}
	
		// call on an existing instance: like creating an instance of a class
	new { |...args| ^this.copy.initialize(*args) }
	
		// init clears the environment
	init { arg func, initEnv, parentKeys;
		var new;
		initEnv.isNil.if({
				// the initfunc needs to establish all the envir vars
			env = Environment.make(func).moveFunctionsToParent(parentKeys);
		}, {
			new = initEnv.shallowCopyItems.make(func);
			new.parent_(initEnv.parent.copy);
			env = new.moveFunctionsToParent(parentKeys);
		});
	}
	
	storeArgs { ^[env] }
	
	at { |key| ^env[key] }
	put { |key, value| env.put(key, value) }
	putAll { |... dictionaries| env.putAll(*dictionaries) }
	
	next { arg ... args;
		var result;
		env.use({ result = ~next.valueArray([this] ++ args); });
		^result
	}
	
	value { arg ... args;
		var result;
		env.use({ result = ~next.valueArray([this] ++ args); });
		^result
	}
	
	reset { arg ... args;
		var result;
		env.use({ result = ~reset.valueArray([this] ++ args); });
		^result
	}
	
	update { arg ... args;
		var result;
		env.use({ result = ~update.valueArray([this] ++ args); });
		^result
	}
	
	asStream { arg ... args;
		var result;
		env.use({ result = ~asStream.valueArray([this] ++ args); });
		^result
	}
	
	asPattern { arg ... args;
		var result;
		env.use({ result = ~asPattern.valueArray([this] ++ args); });
		^result
	}
	
	use { arg func;
		var result;
		env.use({ result = func.value; });
		^result
	}
	
	free { arg ... args;
//		env.put(\me, nil);		// may be necessary for gc to explicitly kill the circular ref.
		env.use({ ~free.valueArray([this] ++ args) });
		env = nil;
	}
	
		// make the AdhocClass act like an object
		// selector must be implemented as a func assigned to an environment var
		// named for the selector
		// if that key returns nil, object does not understand!
		
		// messages not understood by AdhocClass should be passed to the environment
		// so: node.message == node.perform(\message) == node.use({ ~message.value })
	doesNotUnderstand { arg selector ... args;
		var result, item;
		(item = env.at(selector)).isFunction.if({
			env.use({ result = item.valueArray([this] ++ args) });
		}, {
				// if setter is defined as a function in the environment, it will be
				// handled by the above branch -- this needs only to do a simple put
			selector.isSetter.if({
				env.put(selector.asGetter, args[0]);
				result = this
			}, {
				result = item
			});
		});
		^result
	}
	
	perform { arg selector ... args;
//		super.respondsTo(selector).if({
//			^super.performList(selector, args)
//		}, {
			^this.performList(\doesNotUnderstand, [selector] ++ args);
//		});
	}

	tryPerform { arg selector ... args;	// for sth like draggedInto...GUI
		^this.perform(selector, *args);
	}
	
	respondsTo { arg selector;
		^super.respondsTo(selector) or: { env.keys.includes(selector) }
	}
	
		// make a copy of this node, and run this func to change some props
	clone { arg modFunc, parentKeys;
		modFunc.isNil.if({
				// no modification, just return a straight copy
			^this.copy
		}, {
				// create a new AdhocClass, applying the modFunc to the existing environment
			^this.class.new(modFunc, env, parentKeys)
		});
	}

	copy { ^this.shallowCopyItems }
	shallowCopyItems { ^this.class.new
		.env_(env.shallowCopyItems.parent_(env.parent/*.shallowCopy*/))
	}
	
//	copy {
//		var new;
//		new = this.class.new({});
//		env.keysValuesDo({ |k, v| 
//				new.env.put(k,
//					v.isKindOf(Function).if({ v }, { v.copy }));
//		});
//		new.env.parent = Environment.new;
//		env.parent.keysValuesDo({ |k, v| 
//				new.env.parent.put(k,
//					v.isKindOf(Function).if({ v }, { v.copy }));
//		});
//		^new
//	}
//	
	listVars { this.help(\var) }
	listMethods { this.help(\method) }
	
	help { |what = \all|
		(what == \var or: { what == \all }).if({
			"\nVariables:".postln;
			env.parent.reject(_.isFunction)
				.putAll(env.reject(_.isFunction))
				.asSortedArray.do(_.postln);
		});
		(what == \method or: { what == \all }).if({
			"\nMethods:".postln;
			env.parent.select(_.isFunction)
				.putAll(env.select(_.isFunction))
				.asSortedArray.do({ |pair|
					Post << pair[0] << "(";
					pair[1].listArgs;
					Post << ")\n";
				});
		});
	}

}
