#include "SC_PlugIn.h"

// InterfaceTable contains pointers to functions in the host (server).
static InterfaceTable *ft;

 // declare struct to hold unit generator state
struct MyDelay : public Unit
{
	uint32 mDelayLength;
	uint32 mPosition;
	float *mData; // delay buffer
	uint32 lDelayLength;
	uint32 lPosition;
	float *lData; // delay buffer for lowpass filter
};
 

// declare unit generator functions 
extern "C"
{
	void load(InterfaceTable *inTable);
 	
	void MyDelay_next_notfull(MyDelay *unit, int inNumSamples);
	void MyDelay_next_full(MyDelay *unit, int inNumSamples);
	void MyDelay_Ctor(MyDelay* unit);
	void MyDelay_Dtor(MyDelay* unit);

};

//////////////////////////////////////////////////////////////////

// Ctor is called to initialize the unit generator. 
// It only executes once.

// A Ctor usually does 3 things.
// 1. set the calculation function.
// 2. initialize the unit generator state variables.
// 3. calculate one sample of output.
 
 void MyDelay_Ctor(MyDelay* unit)
{

	// 1. set the calculation function.
	SETCALC(MyDelay_next_notfull);

	// 2. initialize the unit generator state variables.
	// get the delay length
	unit->mDelayLength = (uint32)(IN0(1) * SAMPLERATE);
	
	// allocate the buffer
	unit->mData = (float*)RTAlloc(unit->mWorld, unit->mDelayLength * sizeof(float));
	// RTAlloc allocates out of the real time memory pool of the server 
	// which is finite. The size of the real time memory pool is set using the
	// -m command line argument of the server.
 
	// initialize the position
	
	unit->mPosition = 0;
	
	// for lowpass filter

	unit->lDelayLength = (uint32) (2); 
	unit->lData = (float*)RTAlloc(unit->mWorld, unit->lDelayLength * sizeof(float));
 	unit->lPosition = 0;
	
	// 3. calculate one sample of output.
	MyDelay_next_notfull(unit, 1);
}

//////////////////////////////////////////////////////////////////

// Dtor is called to perform any clean up for the unit generator. 

void MyDelay_Dtor(MyDelay* unit)
{
	// free the buffer
	RTFree(unit->mWorld, unit->mData);
}


//////////////////////////////////////////////////////////////////

// The calculation function executes once per control period 
// which is typically 64 samples.

// calculation function when the buffer has not yet been filled
void MyDelay_next_notfull(MyDelay *unit, int inNumSamples)
{
	// get the pointer to the output buffer
	float *out = OUT(0);
	
	// get the pointer to the input buffer
	float *in = IN(0);
	
	// get values from struct and store them in local variables.
	// The optimizer will cause them to be loaded it into a register.
	float *data = unit->mData;
	uint32 length = unit->mDelayLength;
	uint32 position = unit->mPosition;
	bool wrapped = false;
	
	// perform a loop for the number of samples in the control period.
	// If this unit is audio rate then inNumSamples will be 64 or whatever
	// the block size is. If this unit is control rate then inNumSamples will
	// be 1.
	for (int i=0; i < inNumSamples; ++i)
	{
		// get old value in delay line
		float z = data[position];
		// store new value in delay line
		data[position] = in[i];
		
		// see if the position went to the end of the buffer 
		if (++position >= length) {
			position = 0; // go back to beginning
			wrapped = true; // indicate we have wrapped.
			
			// change the calculation function
			// next time, the MyDelay_next_full function will be called
			SETCALC(MyDelay_next_full);
		}
		// if we have not yet wrapped, then z is garbage from the uninitialized 
		// buffer, so output zero. If we have wrapped, then z is a good value.
		out[i] = wrapped ? 0.f : z;
	}

	// store the position back to the struct
	unit->mPosition = position;
}

//////////////////////////////////////////////////////////////////

// calculation function when the buffer has been filled
void MyDelay_next_full(MyDelay *unit, int inNumSamples)
{
	// get the pointer to the output buffer
	float *out = OUT(0);	
	float *in = IN(0);
	float *ldata = unit->lData;
	float *data = unit->mData;
	uint32 length = unit->mDelayLength;
	uint32 position = unit->mPosition;
	uint32 llength = unit->lDelayLength;
	uint32 lposition = unit->lPosition;
 	
	for (int i=0; i < inNumSamples; ++i)
	{
	 
		float z ;
	 
		z = in[i] + (data[position] * 0.99);
		// lowpass
		ldata[lposition] = z;

		if (++lposition >= llength) {
			lposition = 0;
			}; 
		z = 0.5 * (ldata[lposition] + z);
		
 		data[position] = z;

		if (++position >= length) {
			position = 0; // go back to beginning
		};
	 
		
		out[i] = z;
	}

	// store the position back to the struct
	unit->mPosition = position;
}
 ////////////////////////////////////////////////////////////////////
 